@@ -4,149 +4,177 @@
 #include <hdf5.h>
 #include <Rinternals.h>    
 #include <R.h>
+#include <malloc.h>
+
+#define DEBUG 0
+#define MEMORYDEBUG 0
 
-#define DEBUG 1
 #define HID(argname) (((h5_holder*) R_ExternalPtrAddr(argname))->id)
 #define NM(argname) (CHAR(STRING_ELT(argname, 0)))
 #define SUCCESS ScalarLogical(1)
 #define FAILURE ScalarLogical(0)
 
 typedef struct h5_holder {
-  int is_file;
-  hid_t id;
+    int is_file;
+    hid_t id;
 } h5_holder;
 
+SEXP h5R_malloc_trim() {
+#ifdef __GLIBC__
+    return ScalarInteger(malloc_trim(0));
+#else
+    return ScalarInteger(-1);
+#endif
+}
+
 void h5R_finalizer(SEXP h5_obj) {
-  h5_holder* h = (h5_holder*) R_ExternalPtrAddr(h5_obj);
-  if (! h) {
-    return;
-  }
-  if (h->is_file == 1) {
-    H5Fflush(HID(h5_obj), H5F_SCOPE_GLOBAL);
-    H5Fclose(HID(h5_obj));
-  } else {
-    switch (H5Iget_type(HID(h5_obj))) {
-    case H5I_DATASET:
-      H5Dclose(HID(h5_obj));
-      break;
-    case H5I_ATTR:
-      H5Aclose(HID(h5_obj));
-      break;
-    case H5I_GROUP:
-      H5Gclose(HID(h5_obj));
-      break;
-    default:
-      break;
+    h5_holder* h = (h5_holder*) R_ExternalPtrAddr(h5_obj);
+    if (! h) {
+	return;
     }
-  }
-  Free(h);
-  R_ClearExternalPtr(h5_obj);
+    if (h->is_file == 1) {
+	if (MEMORYDEBUG) Rprintf("Flushing and closing file.\n");
+	H5Fflush(HID(h5_obj), H5F_SCOPE_GLOBAL);
+	H5Fclose(HID(h5_obj));
+    } else {
+	switch (H5Iget_type(HID(h5_obj))) {
+	case H5I_DATASET:
+	    if (MEMORYDEBUG) Rprintf("Closing dataset.\n");
+	    H5Dclose(HID(h5_obj));
+	    break;
+	case H5I_ATTR:
+	    if (MEMORYDEBUG) Rprintf("Closing attribute.\n");
+	    H5Aclose(HID(h5_obj));
+	    break;
+	case H5I_GROUP:
+	    if (MEMORYDEBUG) Rprintf("Closing group.\n");
+	    H5Gclose(HID(h5_obj));
+	    break;
+	default:
+	    break;
+	}
+    }
+    Free(h);
+    R_ClearExternalPtr(h5_obj);
+    
+    // This call is seemingly a noop.
+    H5garbage_collect();
+
+    // Not sure if calling this here will be sufficient. 
+    h5R_malloc_trim();
 }
 
 SEXP _h5R_make_holder (hid_t id, int is_file) {
-  if (id < 0) {
-    return R_NilValue;
-  } 
-  h5_holder* holder = (h5_holder*) Calloc(1, h5_holder);
-  holder->id = id;
-  holder->is_file = is_file;
-  SEXP e_ptr = R_MakeExternalPtr(holder, R_NilValue, R_NilValue); 
-  PROTECT(e_ptr);
-  R_RegisterCFinalizerEx(e_ptr, h5R_finalizer, TRUE);
-  UNPROTECT(1); 
-  return e_ptr;
+    if (id < 0) {
+	return R_NilValue;
+    } 
+    h5_holder* holder = (h5_holder*) Calloc(1, h5_holder);
+    holder->id = id;
+    holder->is_file = is_file;
+    SEXP e_ptr = R_MakeExternalPtr(holder, R_NilValue, R_NilValue); 
+    PROTECT(e_ptr);
+    R_RegisterCFinalizerEx(e_ptr, h5R_finalizer, TRUE);
+    UNPROTECT(1); 
+    return e_ptr;
 }
 
 SEXP h5R_flush(SEXP h5_file) {
-  H5Fflush(HID(h5_file), H5F_SCOPE_GLOBAL);
-  return SUCCESS;
+    H5Fflush(HID(h5_file), H5F_SCOPE_GLOBAL);
+    return SUCCESS;
 }
 
 SEXP h5R_open(SEXP filename, SEXP mode) {
-  int _mode_ = (INTEGER(mode)[0] == 1) ? H5F_ACC_RDWR : H5F_ACC_RDONLY;
-  return _h5R_make_holder(H5Fopen(NM(filename), _mode_, H5P_DEFAULT), 1);
+    int _mode_ = (INTEGER(mode)[0] == 1) ? H5F_ACC_RDWR : H5F_ACC_RDONLY;
+    return _h5R_make_holder(H5Fopen(NM(filename), _mode_, H5P_DEFAULT), 1);
 }
 
 SEXP h5R_create(SEXP filename) {
-  return _h5R_make_holder(H5Fcreate(NM(filename), H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT), 1);
+    return _h5R_make_holder(H5Fcreate(NM(filename), H5F_ACC_EXCL, H5P_DEFAULT, H5P_DEFAULT), 1);
 }
 
 SEXP h5R_get_group(SEXP h5_obj, SEXP group_name) {
-  return _h5R_make_holder(H5Gopen2(HID(h5_obj), NM(group_name), H5P_DEFAULT), 0);
+    return _h5R_make_holder(H5Gopen2(HID(h5_obj), NM(group_name), H5P_DEFAULT), 0);
 }
 
 SEXP h5R_create_group(SEXP h5_obj, SEXP group_name) {
-  return _h5R_make_holder(H5Gcreate2(HID(h5_obj), NM(group_name), H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT), 0);
+    return _h5R_make_holder(H5Gcreate2(HID(h5_obj), NM(group_name), H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT), 0);
 }
 
 SEXP h5R_get_dataset(SEXP h5_obj, SEXP dataset_name) {
-  return _h5R_make_holder(H5Dopen2(HID(h5_obj), NM(dataset_name), H5P_DEFAULT), 0);
+    return _h5R_make_holder(H5Dopen2(HID(h5_obj), NM(dataset_name), H5P_DEFAULT), 0);
 }
 
 SEXP h5R_get_attr(SEXP h5_obj, SEXP attr_name) {
-  return _h5R_make_holder(H5Aopen(HID(h5_obj), NM(attr_name), H5P_DEFAULT), 0);
+    return _h5R_make_holder(H5Aopen(HID(h5_obj), NM(attr_name), H5P_DEFAULT), 0);
 }
 
+SEXP h5R_garbage_collect() {
+    return ScalarInteger(H5garbage_collect());
+}
+
+// callers of this function need to clean up using: H5Sclose.
 hid_t _h5R_get_space(SEXP h5_obj) {
-  hid_t space = -1;
-
-  switch (H5Iget_type(HID(h5_obj))) {
-  case H5I_DATASET:
-    space = H5Dget_space(HID(h5_obj));
-    break;
-  case H5I_ATTR:
-    space = H5Aget_space(HID(h5_obj));
-    break;
-  default:
-    error("Unknown object in %s.\n", __func__);
-  }
-  return space;
+    hid_t space = -1;
+    switch (H5Iget_type(HID(h5_obj))) {
+    case H5I_DATASET:
+	space = H5Dget_space(HID(h5_obj));
+	break;
+    case H5I_ATTR:
+	space = H5Aget_space(HID(h5_obj));
+	break;
+    default:
+	error("Unknown object type in %s.\n", __func__);
+    }
+    return space;
 }
 
+// callers of this function need to clean up using: H5Tclose. 
 hid_t _h5R_get_type(SEXP h5_obj) {
-  hid_t dtype = -1;
-
-  switch (H5Iget_type(HID(h5_obj))) {
-  case H5I_DATASET:
-    dtype = H5Dget_type(HID(h5_obj));
-    break;
-  case H5I_ATTR:
-    dtype = H5Aget_type(HID(h5_obj));
-    break;
-  default:
-    error("Unknown object in %s.\n", __func__);
-  }
-  return dtype;
+    hid_t dtype = -1;
+    switch (H5Iget_type(HID(h5_obj))) {
+    case H5I_DATASET:
+	dtype = H5Dget_type(HID(h5_obj));
+	break;
+    case H5I_ATTR:
+	dtype = H5Aget_type(HID(h5_obj));
+	break;
+    default:
+	error("Unknown object type in %s.\n", __func__);
+    }
+    return dtype;
 }
 
 int _h5R_get_size (SEXP h5_obj) {
-  int s = -1;
-
-  switch (H5Iget_type(HID(h5_obj))) {
-  case H5I_DATASET:
-    s = H5Tget_size(H5Dget_type(HID(h5_obj)));
-    break;
-  case H5I_ATTR:
-    s = H5Tget_size(H5Aget_type(HID(h5_obj)));
-    break;
-  default:
-    error("Unknown object in %s.\n", __func__);
-  }
-  return s;
+    int s   = -1;
+    hid_t t = -1; 
+    switch (H5Iget_type(HID(h5_obj))) {
+    case H5I_DATASET:
+	t = H5Dget_type(HID(h5_obj));
+	s = H5Tget_size(t);
+	break;
+    case H5I_ATTR:
+	t = H5Aget_type(HID(h5_obj));
+	s = H5Tget_size(t);
+	break;
+    default:
+	error("Unknown object type in %s.\n", __func__);
+    }
+    H5Tclose(t);
+    return s;
 }
 
 int _h5R_is_vlen (SEXP h5_obj) {
-  hid_t hdty = _h5R_get_type(h5_obj);
-  int rval = H5Tis_variable_str(hdty);
-  H5Tclose(hdty);
-  return rval;
+    hid_t hdty = _h5R_get_type(h5_obj);
+    int rval = H5Tis_variable_str(hdty);
+    H5Tclose(hdty);
+    return rval;
 }
 
 int _h5R_get_ndims(SEXP h5_obj) {
-  hid_t space = _h5R_get_space(h5_obj);
-  int ndims = H5Sget_simple_extent_ndims(space);
-  H5Sclose(space);
-  return ((ndims < 0) ? 1 : ndims);
+    hid_t space = _h5R_get_space(h5_obj);
+    int ndims = H5Sget_simple_extent_ndims(space);
+    H5Sclose(space);
+    return ((ndims < 0) ? 1 : ndims);
 }
 
 int _h5R_get_nelts(SEXP h5_obj) {
@@ -190,6 +218,10 @@ SEXP h5R_get_dims(SEXP h5_obj) {
   return res;
 }
 
+SEXP h5R_get_object_count(SEXP h5_file) {
+    return ScalarInteger(H5Fget_obj_count(HID(h5_file), H5F_OBJ_ALL));
+}
+
 /** ******************************************************************************
  *
  * Read API
@@ -252,7 +284,8 @@ SEXP _h5R_read_string(SEXP h5_obj, int nelts, hid_t memspace, hid_t filespace) {
   UNPROTECT(1); 
 
   if (_h5R_is_vlen(h5_obj)) {
-    // It doesn't like the H5S_ALL space in the vlen_reclaim.
+    // It doesn't like the H5S_ALL space in the vlen_reclaim which can
+    // be passed in.
     if (memspace == H5S_ALL) {
       hid_t space = _h5R_get_space(h5_obj);
       H5Dvlen_reclaim (memtype, space, H5P_DEFAULT, buf);
@@ -268,8 +301,109 @@ SEXP _h5R_read_string(SEXP h5_obj, int nelts, hid_t memspace, hid_t filespace) {
   return dta;
 }
 
+/** XXX: This is probably leaky **/
 SEXP _h5R_read_compound(SEXP h5_obj, int nelts, hid_t memspace, hid_t filespace) {
-  return R_NilValue;
+  hid_t hdty   = _h5R_get_type(h5_obj);
+  int nmembers = H5Tget_nmembers(hdty);
+  int c_size   = 0;
+  SEXP res     = R_NilValue;
+  hid_t memtype, strtype; 
+  int has_str  = -1;
+
+  for (int i = 0; i < nmembers; i++) 
+    c_size += (H5Tget_size(H5Tget_member_type(hdty, i)));
+
+  memtype = H5Tcreate (H5T_COMPOUND, c_size);
+  PROTECT(res = allocVector(VECSXP, nmembers));
+
+  /** construct the memtype so you can read in the data. **/
+  c_size = 0;
+  for (int i = 0; i < nmembers; i++) {
+    switch (H5Tget_member_class(hdty, i)) {
+    case H5T_INTEGER: 
+      H5Tinsert (memtype, H5Tget_member_name(hdty, i), c_size, H5T_NATIVE_INT);
+      SET_VECTOR_ELT(res, i, allocVector(INTSXP, nelts));
+      break;
+    case H5T_FLOAT:
+      H5Tinsert (memtype, H5Tget_member_name(hdty, i), c_size, H5T_NATIVE_DOUBLE);
+      SET_VECTOR_ELT(res, i, allocVector(REALSXP, nelts));
+      break;
+    case H5T_STRING:
+      strtype = H5Tcopy (H5T_C_S1);
+      has_str = 1;
+      if (H5Tis_variable_str(H5Tget_member_type(hdty, i))) {
+	H5Tset_size(strtype, H5T_VARIABLE);
+	H5Tinsert(memtype, H5Tget_member_name(hdty, i), c_size, strtype);
+      } else {
+	H5Tset_size(strtype, H5Tget_size(H5Tget_member_type(hdty, i)));
+	H5Tinsert(memtype, H5Tget_member_name(hdty, i), c_size, strtype);
+      }
+      SET_VECTOR_ELT(res, i, allocVector(STRSXP, nelts));
+      break;
+    default:
+      error("Unsupported class %d in %s\n", H5Tget_member_class(hdty, i), __func__);
+      break;
+    }
+    c_size += (H5Tget_size(H5Tget_member_type(hdty, i)));
+  }
+
+  /** read **/
+  void* buf = malloc(c_size*nelts);
+  H5Dread (HID(h5_obj), memtype, memspace, filespace, H5P_DEFAULT, buf);
+  
+  /** put the results into a list. **/
+  for (int i = 0; i < nelts; i++) {
+    int current_offset = 0;
+    for (int j = 0; j < nmembers; j++) {
+      switch (H5Tget_member_class(hdty, j)) {
+      case H5T_INTEGER: 
+	INTEGER(VECTOR_ELT(res, j))[i] = ((int*) (buf + i*c_size + current_offset))[0];
+	break;
+      case H5T_FLOAT:
+	REAL(VECTOR_ELT(res, j))[i] = ((double*) (buf + i*c_size + current_offset))[0];
+	break;
+      case H5T_STRING:
+	if (H5Tis_variable_str(H5Tget_member_type(hdty, j))) {
+	  SET_STRING_ELT(VECTOR_ELT(res, j), i, mkChar(((char**) (buf + i*c_size + current_offset))[0]));
+	} else {
+	  SET_STRING_ELT(VECTOR_ELT(res, j), i, mkChar(((char*) (buf + i*c_size + current_offset))));
+	}
+	break;
+      default:
+	error("Unsupported class in %s\n", __func__);
+	break;
+      }
+      current_offset += H5Tget_size(H5Tget_member_type(hdty, j));
+    }
+  }
+
+  /** names **/
+  SEXP lst_names;
+
+  PROTECT(lst_names = allocVector(VECSXP, nmembers));
+  for (int j = 0; j < nmembers; j++) {
+    SET_VECTOR_ELT(lst_names, j, mkChar(H5Tget_member_name(hdty, j)));
+  }
+  setAttrib(res, R_NamesSymbol, lst_names);
+  
+  if (has_str > 0) { 
+    /** cleanup **/
+    if (memspace == H5S_ALL) {
+      hid_t space = _h5R_get_space(h5_obj);
+      H5Dvlen_reclaim (memtype, space, H5P_DEFAULT, buf);
+      H5Sclose(space);
+    } else {
+      H5Dvlen_reclaim (memtype, memspace, H5P_DEFAULT, buf);
+    }
+    H5Tclose(strtype);
+  }
+  free(buf);
+  H5Tclose(memtype);
+  H5Tclose(hdty);
+
+  UNPROTECT(2);
+ 
+  return res;
 }
 
 SEXP h5R_read_dataset_all(SEXP h5_dataset) {
@@ -415,8 +549,12 @@ SEXP h5R_read_1d_slabs(SEXP h5_dataset, SEXP _offsets, SEXP _counts) {
     
   for (i = 0; i < rlen; i++) {
     INTEGER(_SEXP_offsets)[0] = offsets[i];
-    INTEGER(_SEXP_counts)[0] = counts[i];
-    SET_VECTOR_ELT(r_lst, i, h5R_read_dataset(h5_dataset, _SEXP_offsets, _SEXP_counts));
+    INTEGER(_SEXP_counts)[0]  = counts[i];
+    if (counts[i] <= 0) {
+	SET_VECTOR_ELT(r_lst, i, R_NilValue);
+    } else {
+	SET_VECTOR_ELT(r_lst, i, h5R_read_dataset(h5_dataset, _SEXP_offsets, _SEXP_counts));
+    }
   }
   UNPROTECT(3);
   return(r_lst);
